/*
 * Copyright (C) 2003
 * 	Hidetoshi Shimokawa. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *
 *	This product includes software developed by Hidetoshi Shimokawa.
 *
 * 4. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: dcons.c,v 1.36 2003/02/12 02:05:56 simokawa Exp $
 * $FreeBSD$
 */

#include <sys/param.h>
#include <sys/kernel.h>
#include <sys/systm.h>
#include <sys/types.h>
#include <sys/conf.h>
#include <sys/cons.h>
#include <sys/consio.h>
#include <sys/tty.h>
#include <sys/malloc.h>
#include <sys/proc.h>
#include <sys/ucred.h>
/* for vtophys */
#include <vm/vm.h>
#include <vm/pmap.h>
#ifdef __alpha__
#undef vtophys
#define vtophys(va)	alpha_XXX_dmamap((vm_offset_t)(va))
#endif /* __alpha__ */

#include <dev/dcons/dcons.h>

#include <ddb/ddb.h>
#include <sys/reboot.h>

#include <sys/sysctl.h>

#include "opt_ddb.h"
#include "opt_comconsole.h"

#define DCONS_POLL_HZ	100
#define FORCE_CONSOLE	0	/* mostly for FreeBSD-4 */
#define FORCE_GDB_DEV	1

#define DCONSCTL	2

#define	CDEV_MAJOR	184
static d_open_t		dcons_open;
static d_close_t	dcons_close;
static d_ioctl_t	dcons_ioctl;

static struct cdevsw dcons_cdevsw = {
	/* open */	dcons_open,
	/* close */	dcons_close,
	/* read */	ttyread,
	/* write */	ttywrite,
	/* ioctl */	dcons_ioctl,
	/* poll */	ttypoll,
	/* mmap */	nommap,
	/* strategy */	nostrategy,
	/* name */	"dcons",
	/* major */	CDEV_MAJOR,
	/* dump */	nodump,
	/* psize */	nopsize,
	/* flags */	0,
};

#define DBUF_SIZE 8192
#ifndef KLD_MODULE
static char dcons_buf[DBUF_SIZE];
#endif
struct dcons_buf *dbuf;
vm_offset_t dcons_paddr;

SYSCTL_NODE(_kern, OID_AUTO, dcons, CTLFLAG_RD, 0, "Dumb Console");
SYSCTL_INT(_kern_dcons, OID_AUTO, paddr, CTLFLAG_RD, &dcons_paddr, 0,
				"Physical address of the dcons buffer");

static int drv_init = 0;
static int polltime;
static struct callout dcons_callout;

/* per device data */
static struct dcons_softc {
	dev_t dev;
	int polltime;
	struct callout dcons_callout;
	struct dcons_ch	o, i;
	int brk_state;
} sc[DCONS_NPORT];
static void	dcons_tty_start(struct tty *);
static int	dcons_tty_param(struct tty *, struct termios *);
static void	dcons_timeout(void *);
static int	dcons_drv_init(int);

static cn_probe_t	dcons_cnprobe;
static cn_init_t	dcons_cninit;
static cn_getc_t	dcons_cngetc;
static cn_checkc_t 	dcons_cncheckc;
static cn_putc_t	dcons_cnputc;

CONS_DRIVER(dcons, dcons_cnprobe, dcons_cninit, NULL, dcons_cngetc,
    dcons_cncheckc, dcons_cnputc, NULL);

#if __FreeBSD_version < 500000
#define THREAD	proc
#else
#define THREAD	thread
#endif

static int
dcons_open(dev_t dev, int flag, int mode, struct THREAD *td)
{
	struct tty *tp;
	struct dcons_softc *dc;
	int unit, error, s;

	unit = minor(dev);
	if (unit != 0 && unit != DCONSCTL) {
		return (ENXIO);
	}

	tp = dev->si_tty = ttymalloc(dev->si_tty);
	tp->t_oproc = dcons_tty_start;
	tp->t_param = dcons_tty_param;
	tp->t_stop = nottystop;
	tp->t_dev = dev;

	error = 0;
	dc = (struct dcons_softc *)dev->si_drv1;

	s = spltty();
	if ((tp->t_state & TS_ISOPEN) == 0) {
		tp->t_state |= TS_CARR_ON;
		ttychars(tp);
		tp->t_iflag = TTYDEF_IFLAG;
		tp->t_oflag = TTYDEF_OFLAG;
		tp->t_cflag = TTYDEF_CFLAG|CLOCAL;
		tp->t_lflag = TTYDEF_LFLAG;
		tp->t_ispeed = tp->t_ospeed = TTYDEF_SPEED;
		ttsetwater(tp);
	} else if ((tp->t_state & TS_XCLUDE) && suser(td)) {
		splx(s);
		return (EBUSY);
	}
	splx(s);

	error = (*linesw[tp->t_line].l_open)(dev, tp);

	return (error);
}

static int
dcons_close(dev_t dev, int flag, int mode, struct THREAD *td)
{
	int	unit;
	struct	tty *tp;

	unit = minor(dev);
	if (unit != 0 && unit != DCONSCTL) {
		return (ENXIO);
	}

	tp = dev->si_tty;
	if (tp->t_state & TS_ISOPEN) {
		struct dcons_softc *dc;

		dc = (struct dcons_softc *)dev->si_drv1;
		(*linesw[tp->t_line].l_close)(tp, flag);
		ttyclose(tp);
	}

	return (0);
}

static int
dcons_ioctl(dev_t dev, u_long cmd, caddr_t data, int flag, struct THREAD *td)
{
	int	unit;
	struct	tty *tp;
	int	error;

	unit = minor(dev);
	if (unit != 0 && unit != DCONSCTL) {
		return (ENXIO);
	}

	tp = dev->si_tty;
	error = (*linesw[tp->t_line].l_ioctl)(tp, cmd, data, flag, td);
	if (error != ENOIOCTL) {
		return (error);
	}

	error = ttioctl(tp, cmd, data, flag);
	if (error != ENOIOCTL) {
		return (error);
	}

	return (ENOTTY);
}

static int
dcons_tty_param(struct tty *tp, struct termios *t)
{
	tp->t_ispeed = t->c_ispeed;
	tp->t_ospeed = t->c_ospeed;
	tp->t_cflag = t->c_cflag;
	return 0;
}

static void
dcons_tty_start(struct tty *tp)
{
	int s;

	s = spltty();
	if (tp->t_state & (TS_TIMEOUT | TS_TTSTOP)) {
		ttwwakeup(tp);
		return;
	}

	tp->t_state |= TS_BUSY;
	while (tp->t_outq.c_cc != 0) {
		dcons_cnputc(tp->t_dev, getc(&tp->t_outq));
	}
	tp->t_state &= ~TS_BUSY;

	ttwwakeup(tp);
	splx(s);
}

static void
dcons_timeout(void *v)
{
	struct	tty *tp;
	struct dcons_softc *dc;
	int i, c;

	for (i = 0; i < DCONS_NPORT; i ++) {
		dc = &sc[i];
		tp = dc->dev->si_tty;
		while ((c = dcons_cncheckc(tp->t_dev)) != -1) {
			if (tp->t_state & TS_ISOPEN) {
				(*linesw[tp->t_line].l_rint)(c, tp);
			}
		}
	}
	callout_reset(&dc->dcons_callout, dc->polltime, dcons_timeout, tp);
}

static void
dcons_cnprobe(struct consdev *cp)
{
	cp->cn_dev = makedev(CDEV_MAJOR, DCONSCTL);
#if FORCE_CONSOLE
	cp->cn_pri = CN_REMOTE;
#else
	cp->cn_pri = CN_NORMAL;
#endif
}

static void
dcons_cninit(struct consdev *cp)
{
	dcons_drv_init(0);
	cp->cn_dev->si_drv1 = (void *)&sc[DCONS_CON];
				/* share port0 with unit0 */
}

static int
dcons_cngetc(dev_t dev)
{
	int c;

	while ((c = dcons_cncheckc(dev)) == -1);

	return (c & 0xff);
}

static int
dcons_cncheckc(dev_t dev)
{
	unsigned char c;
	u_int32_t ptr, pos, gen, next_gen;
	struct dcons_softc *dc;
	struct dcons_ch *ch;

	dc = (struct dcons_softc *)dev->si_drv1;
	ch = &dc->i;

	ptr = ntohl(*ch->ptr);
	gen = ptr >> DCONS_GEN_SHIFT;
	pos = ptr & DCONS_POS_MASK;
	if (gen == ch->gen && pos == ch->pos)
		return (-1);

	next_gen = DCONS_NEXT_GEN(ch->gen);
	/* XXX sanity check */
	if ((gen != ch->gen && gen != next_gen)
			|| (gen == ch->gen && pos < ch->pos)) {
		/* generation skipped !! */
		/* XXX discard */
		ch->gen = gen;
		ch->pos = pos;
		return (-1);
	}

	c = ch->buf[ch->pos];
	ch->pos ++;
	if (ch->pos >= ch->size) {
		ch->gen = next_gen;
		ch->pos = 0;
	}

#if DDB && ALT_BREAK_TO_DEBUGGER
#define KEY_CTRLB	2	/* ^B */
#define KEY_CR		13	/* CR '\r' */
#define KEY_TILDE	126	/* ~ */
#define STATE0		0
#define STATE1		1
#define STATE2		2
	switch (dc->brk_state) {
	case STATE0:
		if (c == KEY_CR)
			dc->brk_state = STATE1;
		break;
	case STATE1:
		if (c == KEY_TILDE)
			dc->brk_state = STATE2;
		else
			dc->brk_state = STATE0;
		break;
	case STATE2:
		dc->brk_state = STATE0;
		if (c == KEY_CTRLB) {
#if FORCE_GDB_DEV
			if (dev == gdbdev)
				boothowto |= RB_GDB;
#endif
			breakpoint();
		}
	}
#endif
	return (c);
}

static void
dcons_cnputc(dev_t dev, int c)
{
	struct dcons_softc *dc;
	struct dcons_ch *ch;

	dc = (struct dcons_softc *)dev->si_drv1;
	ch = &dc->o;

	ch->buf[ch->pos] = c;
	ch->pos ++;
	if (ch->pos >= ch->size) {
		ch->gen = DCONS_NEXT_GEN(ch->gen);
		ch->pos = 0;
	}
	*ch->ptr = DCONS_MAKE_PTR(ch);
}

static int
dcons_init_port(int port, int offset, int size)
{
	int osize;
	struct dcons_softc *dc;

	dc = &sc[port];

	osize = size * 3 / 4;

	dc->o.size = osize;
	dc->i.size = size - osize;
	dc->o.buf = (char *)dbuf + offset;
	dc->i.buf = dc->o.buf + osize;
	dc->o.gen = dc->i.gen = 0;
	dc->o.pos = dc->i.pos = 0;
	dc->o.ptr = &dbuf->optr[port];
	dc->i.ptr = &dbuf->iptr[port];
	dc->brk_state = STATE0;
	dbuf->osize[port] = htonl(osize);
	dbuf->isize[port] = htonl(size - osize);
	dbuf->ooffset[port] = htonl(offset);
	dbuf->ioffset[port] = htonl(offset + osize);
	dbuf->optr[port] = DCONS_MAKE_PTR(&dc->o);
	dbuf->iptr[port] = DCONS_MAKE_PTR(&dc->i);

	return(0);
}

static int
dcons_drv_init(int stage)
{
	int buf_size=DBUF_SIZE, size, size0, offset;

	if (drv_init)
		return(drv_init);

	drv_init = -1;

#ifndef KLD_MODULE
	if (stage == 0)
		/*
		 * FORCE_CONSOLE == 1 and statically linked.
		 * called from cninit(). can't use contigmalloc yet .
		 */
		dbuf = (struct dcons_buf *) dcons_buf;
	else
#endif
		/*
		 * FORCE_CONSOLE == 0 or kernel module case.
		 * if the module is loaded after boot,
		 * dcons_buf could be non-continuous.
		 */ 
		dbuf = (struct dcons_buf *) contigmalloc(buf_size,
			M_DEVBUF, 0, 0x10000, 0xffffffff, PAGE_SIZE, 0ul);

	offset = DCONS_HEADER_SIZE;
	size = (buf_size - offset);
	size0 = size * 3 / 4;

	dcons_init_port(0, offset, size0);
	offset += size0;
	dcons_init_port(1, offset, size - size0);
	dbuf->version = htonl(DCONS_VERSION);
	dbuf->magic = ntohl(DCONS_MAGIC);

#if DDB && FORCE_GDB_DEV 
	gdbdev = makedev(CDEV_MAJOR, DCONS_GDB);
	gdb_getc = dcons_cngetc;
	gdb_putc = dcons_cnputc;
#endif
	drv_init = 1;

	return 0;
}


static int
dcons_attach_port(int port, char *name)
{
	struct dcons_softc *dc;
	struct tty *tp;

	dc = &sc[port];
	dc->dev = make_dev(&dcons_cdevsw, port,
			UID_ROOT, GID_WHEEL, 0600, name);
	tp = ttymalloc(NULL);

	dc->dev->si_drv1 = (void *)dc;
	dc->dev->si_tty = tp;

	tp->t_oproc = dcons_tty_start;
	tp->t_param = dcons_tty_param;
	tp->t_stop = nottystop;
	tp->t_dev = dc->dev;

#if __FreeBSD_version < 500000
	callout_init(&dc->dcons_callout);
#else
	callout_init(&dc->dcons_callout, 0);
#endif

	return(0);
}

static int
dcons_attach(dev_t *dconsctldev)
{

	dcons_paddr = vtophys(dbuf);
#if __FreeBSD_version < 500000
	printf("dcons: virtual %p physical 0x%x quad 0x%x\n",
#else
	printf("dcons: virtual %p physical 0x%tx quad 0x%tx\n",
#endif
			dbuf, dcons_paddr, dcons_paddr>>2);
	dcons_attach_port(DCONS_CON, "dcons");
	dcons_attach_port(DCONS_GDB, "dgdb");
	*dconsctldev = make_dev(&dcons_cdevsw, DCONSCTL,
			UID_ROOT, GID_WHEEL, 0600, "dconsctl");
#if __FreeBSD_version < 500000
	callout_init(&dcons_callout);
#else
	callout_init(&dcons_callout, 0);
#endif
	polltime = hz / DCONS_POLL_HZ;
	if (polltime < 1)
		polltime = 1;
	callout_reset(&dcons_callout, polltime, dcons_timeout, NULL);
	return(0);
}

static int
dcons_detach(int port)
{
	struct	tty *tp;
	struct dcons_softc *dc;

	dc = &sc[port];

	tp = dc->dev->si_tty;

	if (tp->t_state & TS_ISOPEN) {
		printf("dcons: still opened\n");
		(*linesw[tp->t_line].l_close)(tp, 0);
		tp->t_gen++;
		ttyclose(tp);
		ttwakeup(tp);
		ttwwakeup(tp);
	}
	/* XXX
	 * must wait until all device are closed.
	 */
	destroy_dev(dc->dev);

	return(0);
}


/* cnXXX works only for FreeBSD-5 */
static int
dcons_modevent(module_t mode, int type, void *data)
{
	int err = 0;
	static dev_t dconsctldev;

	switch (type) {
	case MOD_LOAD:
		err = dcons_drv_init(1);
#if __FreeBSD_version >= 500000
		if (err == 0) {
			dcons_cnprobe(&dcons_consdev);
			dcons_cninit(&dcons_consdev);
			cnadd(&dcons_consdev);
		}
#endif
		err = dcons_attach(&dconsctldev);
		break;
	case MOD_UNLOAD:
		callout_stop(&dcons_callout);
#if DDB && FORCE_GDB_DEV
		gdbdev = NODEV;
#endif
#if __FreeBSD_version >= 500000
		cnremove(&dcons_consdev);
#endif
		dcons_detach(DCONS_CON);
		dcons_detach(DCONS_GDB);
		destroy_dev(dconsctldev);
		dbuf->magic = 0;
		free(sc, M_DEVBUF);
		contigfree(dbuf, DBUF_SIZE, M_DEVBUF);
		break;
	}
	return(err);
}

DEV_MODULE(dcons, dcons_modevent, NULL);
