/*
 * Copyright (C) 2003
 * 	Hidetoshi Shimokawa. All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *
 *	This product includes software developed by Hidetoshi Shimokawa.
 *
 * 4. Neither the name of the author nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 * $Id: fwchat.c,v 1.34 2003/02/12 02:00:36 simokawa Exp $
 * $FreeBSD$
 */

#include <sys/param.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <dev/dcons/dcons.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <err.h>
#include <string.h>
#include <sys/event.h>
#include <sys/time.h>
#include <arpa/telnet.h>

#include <sys/ioccom.h>
#include <dev/firewire/firewire.h>

#define	DCONS_POLL_HZ	100

int verbose = 0;
int raw = 0;
int alt_break = 0;

static struct dcons_state {
	int fd;
	int kq;
	off_t paddr;
	int fw_ready;
	struct dcons_port {
		int port;
		struct dcons_ch o;
		struct dcons_ch i;
		u_int32_t optr;
		int s;
		int infd;
		int outfd;
		struct addrinfo *res;
		int skip_read;
	} port[DCONS_NPORT];
	struct timespec to;
	struct timespec zero;
} sc;

static int
fwchat_fetch_header(struct dcons_state *dc)
{
	struct dcons_buf dbuf;
	int j;

	if (pread(dc->fd, &dbuf, DCONS_HEADER_SIZE, dc->paddr) < 0) {
		warn("read faild");
		return (-1);
	}
	if (dbuf.magic != htonl(DCONS_MAGIC)) {
		warnx("wrong magic");
		return (-1);
	}
	if (ntohl(dbuf.version) != DCONS_VERSION) {
#if __FreeBSD_version < 500000
		warnx("wrong version %ld,%d",
#else
		warnx("wrong version %d,%d",
#endif
			ntohl(dbuf.version), DCONS_VERSION);
		return (-1);
	}

	for (j = 0; j < DCONS_NPORT; j++) {
		struct dcons_ch *o, *i;
		off_t newbuf;
		int new = 0;

		o = &dc->port[j].o;
		newbuf = dc->paddr + ntohl(dbuf.ooffset[j]);
		o->size = ntohl(dbuf.osize[j]);

		if (newbuf != o->buf) {
			new = 1;
			o->gen = ntohl(dbuf.optr[j]) >> DCONS_GEN_SHIFT;
			o->pos = ntohl(dbuf.optr[j]) & DCONS_POS_MASK;
			o->buf = newbuf;
		}

		i = &dc->port[j].i;
		i->size = ntohl(dbuf.isize[j]);
		i->gen = ntohl(dbuf.iptr[j]) >> DCONS_GEN_SHIFT;
		i->pos = ntohl(dbuf.iptr[j]) & DCONS_POS_MASK;
		i->buf = dc->paddr + ntohl(dbuf.ioffset[j]);

		if (verbose) {
			printf("port %d   size offset   gen   pos\n", j);
#if __FreeBSD_version < 500000
			printf("output: %5d %6ld %5d %5d\n",
#else
			printf("output: %5d %6d %5d %5d\n",
#endif
			o->size, ntohl(dbuf.ooffset[j]), o->gen, o->pos);
#if __FreeBSD_version < 500000
			printf("input : %5d %6ld %5d %5d\n",
#else
			printf("input : %5d %6d %5d %5d\n",
#endif
			i->size, ntohl(dbuf.ioffset[j]), i->gen, i->pos);
		}
#if 1
		if (j == 0 && new) {
			if (o->gen > 0)
				o->gen --;
			else
				o->pos = 0;
		}
#endif
	}
	return(0);
}

static int
fwchat_get_ptr (struct dcons_state *dc) {
	int dlen, i;
	u_int32_t optr[DCONS_NPORT+1];
	static int failed = 0;

	dlen = pread(dc->fd, &optr, sizeof(optr),
		dc->paddr + __offsetof(struct dcons_buf, magic));

	if (dlen < 0) {
		warnx("pread error 1");
		if (failed ++ > 10) {
			dc->fw_ready = 0;
			failed = 0;
		}
		return(-1);
	}
	if (optr[0] != htonl(DCONS_MAGIC)) {
		warnx("wrong magic");
		if (failed ++ > 10) {
			dc->fw_ready = 0;
			failed = 0;
		}
		return(-1);
	}
	failed = 0;
	for (i = 0; i < DCONS_NPORT; i ++)
		dc->port[i].optr = ntohl(optr[i+1]);
	return(0);
}

#define MAX_XFER 2048
static int
fwchat_read_fw(struct dcons_state *dc, int port, char *buf, int len)
{
	struct dcons_ch *ch;
	u_int32_t ptr, pos, gen, next_gen;
	int rlen, dlen, lost;

	ch = &dc->port[port].o;
	ptr = dc->port[port].optr;
	gen = ptr >> DCONS_GEN_SHIFT;
	pos = ptr & DCONS_POS_MASK;
	if (gen == ch->gen && pos == ch->pos)
		return (-1);

	next_gen = DCONS_NEXT_GEN(ch->gen);
	/* XXX sanity check */
	if (gen == ch->gen) {
		if (pos > ch->pos)
			goto ok;
		lost = ch->size * DCONS_GEN_MASK - ch->pos;
		ch->pos = 0;
	} else if (gen == next_gen) {
		if (pos <= ch->pos)
			goto ok;
		lost = pos - ch->pos;
		ch->pos = pos;
	} else {
		lost = gen - ch->gen;
		if (lost < 0)
			lost += DCONS_GEN_MASK;
		if (verbose)
			printf("[genskip %d]", lost);
		lost = lost * ch->size - ch->pos;
		ch->pos = 0;
		ch->gen = gen;
	}
	/* generation skipped !! */
	/* XXX discard */
	if (verbose)
		printf("[lost %d]", lost);
ok:
	if (gen > ch->gen)
		rlen = ch->size - ch->pos;
	else
		rlen = pos - ch->pos;

	if (rlen > MAX_XFER)
		rlen = MAX_XFER;
	if (rlen > len)
		rlen = len;

#if 1
	if (verbose)
		printf("[%d]", rlen); fflush(stdout);
#endif

	dlen = pread(dc->fd, buf, rlen, ch->buf + ch->pos);
	if (dlen < 0) {
		warnx("pread error 2");
		dc->fw_ready = 0;
		return(-1);
	}
	if (dlen != rlen)
		warnx("dlen(%d) != rlen(%d)\n", dlen, rlen);
	ch->pos += dlen;
	if (ch->pos >= ch->size) {
		ch->gen = next_gen;
		ch->pos = 0;
	}
	return (dlen);
}
static int
fwchat_putc(struct dcons_state *dc, int port, int c)
{
	struct dcons_ch *ch;
	char cbuf;
	u_int32_t ptr;
	int len;

	ch = &dc->port[port].i;
	cbuf = c;

	len = pwrite(dc->fd, &cbuf, 1, ch->buf + ch->pos);
	if (len < 0) {
		warnx("pwrite error 1");
		dc->fw_ready = 0;
		return(-1);
	}
	ch->pos ++;
	if (ch->pos >= ch->size) {
		ch->gen = DCONS_NEXT_GEN(ch->gen);
		ch->pos = 0;
	}
	ptr = DCONS_MAKE_PTR(ch);

	if (verbose > 2)
		printf("iptr: 0x%x\n", ptr);
	len = pwrite(dc->fd, &ptr, sizeof(u_int32_t),
		dc->paddr + __offsetof(struct dcons_buf, iptr[port]));
	if (len < 0) {
		warnx("pwrite error 2");
		dc->fw_ready = 0;
		return(-1);
	}
	return(0);
}

static int
fwchat_write_socket(int fd, char *buf, int len)
{
	write(fd, buf, len);
	if (verbose > 1) {
		buf[len] = 0;
		printf("[%s]", buf);
	}
	return (0);
}

static int
fwchat_init_socket(struct dcons_state *dc, int port, char *host, int sport)
{
	struct addrinfo hints, *res;
	int on = 1, error;
	char service[10];
	struct kevent kev;
	struct dcons_port *p;

	p = &dc->port[port];
	p->port = port;
	p->infd = p->outfd = -1;
	memset(&hints, 0, sizeof(hints));
	hints.ai_flags = AI_PASSIVE;
#if 1	/* gdb can talk v4 onyl */
	hints.ai_family = PF_INET;
#else
	hints.ai_family = PF_UNSPEC;
#endif
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_protocol = 0;

	if (verbose)
		printf("%s:%d for port %d\n",
			host == NULL ? "*" : host, sport, port);
	snprintf(service, sizeof(service), "%d", sport);
	error = getaddrinfo(host, service,  &hints, &res);
	if (error)
		errx(1, "tcp/%s: %s\n", service, gai_strerror(error));
	p->res = res;
	p->s = socket(res->ai_family, res->ai_socktype, res->ai_protocol);
	if (p->s < 0)
		err(1, "socket");
	setsockopt(p->s, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on));

	if (bind(p->s, p->res->ai_addr, p->res->ai_addrlen) < 0) {
		err(1, "bind");
	}
	if (listen(p->s, 1) < 0)
		err(1, "listen");
	EV_SET(&kev, p->s, EVFILT_READ, EV_ADD | EV_ONESHOT, 0, 0,
		(void *)((1<<16) | port));
	error = kevent(dc->kq, &kev, 1, NULL, 0, &dc->to);
	if (error < 0)
		err(1, "kevent");
	return p->s;
}

static int
fwchat_accept_socket(struct dcons_state *dc, struct dcons_port *p, int telnet)
{
	int foo, ns, flags;
	struct kevent kev;

	/* accept connection */
	foo = p->res->ai_addrlen;
	ns = accept(p->s, p->res->ai_addr, &foo);
	if (ns < 0)
		err(1, "accept");
	if (verbose)
		printf("port%d accepted\n", telnet ? 0 : 1);

	flags = fcntl(ns, F_GETFL, 0);
	flags |= O_NDELAY;
	fcntl(ns, F_SETFL, flags);
#if 1
	if (telnet && !raw) {
		char sga[] = {IAC, WILL, TELOPT_SGA};
		char linemode[] = {IAC, DONT, TELOPT_LINEMODE};
		char echo[] = {IAC, WILL, TELOPT_ECHO};

		write(ns, sga, sizeof(sga));
		write(ns, linemode, sizeof(linemode));
		write(ns, echo, sizeof(echo));
		p->skip_read = 0;
	}
#endif

	p->infd = p->outfd = ns;
	EV_SET(&kev, ns, EVFILT_READ, EV_ADD,
				NOTE_LOWAT, 1, (void *)p->port);
	kevent(dc->kq, &kev, 1, NULL, 0, &dc->zero);
	return(0);
}

static int
fwchat_read_socket(struct dcons_state *dc, struct dcons_port *p, int telnet)
{
	struct kevent kev;
	int len, i;
	u_char ch, *buf;
	static u_char abreak[3] = {13 /* CR */, 126 /* ~ */, 2 /* ^B */};

	if ((len = read(p->infd, &ch, 1)) > 0) {
		buf = &ch;
		if (telnet && !raw) {
			/* XXX Telent workarounds */
			if (p->skip_read -- > 0)
				return(0);
			if (ch == IAC) {
				if (verbose)
					printf("(IAC)");
				p->skip_read = 2;
				return (0);
			}
			if (ch == 0) {
				if (verbose)
					printf("(0 stripped)");
				return (0);
			}
		} else if (!telnet) {
			/* GDB: ^C -> CR+~+^B */
			if (ch == 0x3 && alt_break) {
				buf = abreak;
				len = sizeof(abreak);
			}
		}
			
		/* XXX discard if not ready*/
		if (dc->fw_ready) {
			for (i = 0; i < len; i ++) {
				fwchat_putc(dc, p->port, buf[i]);
				if (verbose) {
					printf("(%d)", buf[i]);
					fflush(stdout);
				}
			}
		}
	} else {
		if (verbose) {
			if (len == 0)
				warnx("port%d: closed", p->port);
			else
				warn("port%d: read", p->port);
		}
		EV_SET(&kev, p->infd, EVFILT_READ,
			EV_DELETE, 0, 0, NULL);
		kevent(dc->kq, &kev, 1, NULL, 0, &dc->zero);
		close(p->infd);
		close(p->outfd);
		EV_SET(&kev, p->s, EVFILT_READ,
				EV_ADD | EV_ONESHOT, 0, 0,
				(void *)((1<<16) | p->port));
		kevent(dc->kq, &kev, 1, NULL, 0, &dc->zero);
		p->infd = p->outfd = -1;
	}
	return(0);
}
#define NEVENT 5
static int
fwchat_proc_socket(struct dcons_state *dc)
{
	struct kevent elist[NEVENT], *e;
	int i, n, port, ac;
	int telnet;
	struct dcons_port *p;

	n = kevent(dc->kq, NULL, 0, elist, NEVENT, &dc->to);
	for (i = 0; i < n; i ++) {
		e = &elist[i];
		ac = (int)e->udata & (1<<16);
		port = (int)e->udata & ((1<<16) - 1);
		p = &dc->port[port];
		telnet = (port == 0) ? 1 : 0;
		if (ac) {
			fwchat_accept_socket(dc, p, telnet);
		} else {
			fwchat_read_socket(dc, p, telnet);
		}
	}
	return(0);
}

static int
fwchat_proc_fw(struct dcons_state *dc)
{
	int port, len, err;
	char buf[MAX_XFER];
	struct dcons_port *p;

	err = fwchat_get_ptr(dc);
	if (err) {
		/* XXX we should stop write operation too. */
		return err;
	}
	for (port = 0; port < DCONS_NPORT; port ++) {
		p = &dc->port[port];
		if (p->infd < 0)
			continue;
		while ((len = fwchat_read_fw(dc, port, buf, sizeof(buf))) > 0) {
			fwchat_write_socket(p->outfd, buf, len);
			fwchat_get_ptr(dc);
		}
	}
	return 0;
}

static int
fwchat_start_session(struct dcons_state *dc)
{
	int counter = 0;

	while (1) {
		if (!dc->fw_ready && (++counter % 200 == 0))
			dc->fw_ready = (fwchat_fetch_header(dc) == 0);
		if (dc->fw_ready)
			 fwchat_proc_fw(dc);
		fwchat_proc_socket(dc);
	}
	return (0);
}

static void
usage(void)
{
	fprintf(stderr, "usage: fwchat [-brv] [-p base_port]"
				" -t target_eui64 -a address\n");
	fprintf(stderr, "\t-b	translate ctrl-C to CR+~+ctrl-B on gdb port\n");
	fprintf(stderr, "\t-r	disable telnet protocol workaround"
							" on console port\n");
	fprintf(stderr, "\t-v	verbose\n");
	fprintf(stderr, "\t-p	baes inet port address to listen\n");
	fprintf(stderr, "\t-t	EUI64 of target host\n");
	fprintf(stderr, "\t-a	physical address of dcons buffer"
							" on target host\n");
	exit(0);
}
int
main(int argc, char **argv)
{
	struct dcons_state *dc;
	struct fw_eui64 eui;
	int i, ch, error, base_port=5555;
	u_int64_t target = 0;

	dc = &sc;
	dc->paddr = 0;

	while ((ch = getopt(argc, argv, "a:bp:rt:v")) != -1) {
		switch(ch) {
		case 'a':
			dc->paddr = strtoull(optarg, NULL, 0);
			break;
		case 'b':
			alt_break = 1;
			break;
		case 'p':
			base_port = strtoul(optarg, NULL, 0);
			break;
		case 't':
			target = strtoull(optarg, NULL, 0);
			eui.hi = target >> 32;
			eui.lo = target & (((u_int64_t)1 << 32) - 1);
			break;
		case 'v':
			verbose ++;
			break;
		case 'r':
			raw = 1;
			break;
		default:
			usage();
		}
	}
	if (target == 0) {
		warnx("no target specified");
		usage();
	}
	if (dc->paddr == 0) {
		warnx("no address specified");
		usage();
	}

	/* init firewire */
	dc->fw_ready = 0;
	dc->fd = open("/dev/fwmem0", O_RDWR);
	if (dc->fd < 0)
		err(1, "open");
	error = ioctl(dc->fd, FW_SDEUI64, &eui);
	if (error)
		err(1, "ioctl");
	dc->fw_ready = (fwchat_fetch_header(dc) == 0);

	/* iniit sockets */
	dc->kq = kqueue();
	dc->to.tv_sec = 0;
	dc->to.tv_nsec = 1000 * 1000 * 1000 / DCONS_POLL_HZ;
	dc->zero.tv_sec = 0;
	dc->zero.tv_nsec = 0;
	for (i = 0; i < DCONS_NPORT; i++)
		fwchat_init_socket(dc, i, "localhost", base_port + i);

	fwchat_start_session(dc);

	for (i = 0; i < DCONS_NPORT; i++) {
		freeaddrinfo(dc->port[i].res);
	}
	return (0);
}
